import React, { Component, Children, createRef } from 'react';
import PropTypes from 'prop-types';

import styles from './styles.less';

class ReactJSSimpleCarousel extends Component {
  static propTypes = {
    children: PropTypes.node,

    activeSlideIndex: PropTypes.number.isRequired,
    onRequestChange: PropTypes.func.isRequired,
    onAfterChange: PropTypes.func,

    itemsToShow: PropTypes.number,
    itemsToScroll: PropTypes.number,

    updateOnItemClick: PropTypes.bool,

    speed: PropTypes.number,
    delay: PropTypes.number,
    easing: PropTypes.string,

    autoplay: PropTypes.bool,
    autoplayDirection: PropTypes.oneOf(['forward', 'backward']),

    hideNavIfAllVisible: PropTypes.bool,

    containerProps: PropTypes.objectOf(PropTypes.any),
    innerProps: PropTypes.objectOf(PropTypes.any),
    itemsListProps: PropTypes.objectOf(PropTypes.any),
    activeSlideProps: PropTypes.objectOf(PropTypes.any),
    forwardBtnProps: PropTypes.objectOf(PropTypes.any),
    backwardBtnProps: PropTypes.objectOf(PropTypes.any),

    responsiveProps: PropTypes.arrayOf(PropTypes.object),
  };

  static defaultProps = {
    onAfterChange: null,
    children: null,
    speed: 0,
    delay: 0,
    easing: 'linear',
    updateOnItemClick: false,
    autoplay: false,
    autoplayDirection: 'forward',

    itemsToShow: 0,
    itemsToScroll: 1,
    hideNavIfAllVisible: true,

    containerProps: {},
    innerProps: {},
    itemsListProps: {},
    activeSlideProps: {},
    forwardBtnProps: {},
    backwardBtnProps: {},

    responsiveProps: [],
  };

  constructor(props) {
    super(props);

    this.containerRef = createRef();
    this.innerRef = createRef();
    this.itemsListRef = createRef();

    this.resizeTimer = null;
    this.autoplayTimer = null;
    this.itemsListDragStartPos = null;

    this.direction = '';
    this.slides = [];

    this.state = {
      windowWidth: 0,
      positionIndex: props.activeSlideIndex,
    };
  }

  componentDidMount() {
    this.handleInitializationEnd();

    window.addEventListener('resize', this.handleWindowResize);
  }

  componentDidUpdate(prevProps, prevState) {
    const { activeSlideIndex: prevActiveSlideIndex } = this.getRenderProps(prevState, prevProps);
    const { activeSlideIndex } = this.renderProps;

    if (activeSlideIndex === prevActiveSlideIndex) {
      this.direction = '';
    }
  }

  componentWillUnmount() {
    this.stopAutoplay();

    clearTimeout(this.resizeTimer);
    window.removeEventListener('resize', this.handleWindowResize);

    document.removeEventListener('mousemove', this.handleItemsListMouseMove);
    document.removeEventListener('mouseup', this.handleItemsListMouseUp);
    document.removeEventListener('touchmove', this.handleItemsListTouchMove);
    document.removeEventListener('touchend', this.handleItemsListTouchEnd);
  }

  getRenderProps = (state, props) => {
    const { windowWidth } = state;

    const { responsiveProps, ...restProps } = props;

    if (!windowWidth) {
      return restProps;
    }

    const propsByWindowWidth = responsiveProps.reduce((
      result,
      { minWidth = 0, maxWidth = null, ...item } = {},
    ) => {
      if (windowWidth > minWidth && (!maxWidth || windowWidth <= maxWidth)) {
        return {
          ...result,
          ...item,
        };
      }

      return result;
    }, restProps);

    const slidesCount = Children.toArray(propsByWindowWidth.children).length;

    return {
      ...propsByWindowWidth,
      activeSlideIndex: Math.max(
        0,
        Math.min(
          propsByWindowWidth.activeSlideIndex,
          slidesCount - 1,
        ),
      ),
      itemsToShow: Math.min(slidesCount, propsByWindowWidth.itemsToShow),
      itemsToScroll: Math.min(slidesCount, propsByWindowWidth.itemsToScroll),
    };
  }

  getInnerWidth = () => {
    const { windowWidth } = this.state;
    const { itemsToShow, activeSlideIndex } = this.renderProps;

    if (!windowWidth || !itemsToShow) {
      return null;
    }

    return this.slides.reduce((result, item, index) => {
      const isItemVisible = (index >= activeSlideIndex && index < activeSlideIndex + itemsToShow)
        || (
          index < activeSlideIndex
          && (
            index < (activeSlideIndex + itemsToShow - this.slides.length)
          )
        );

      if (!isItemVisible) {
        return result;
      }

      return result + item.current.offsetWidth;
    }, 0);
  }

  getItemsListOffsetBySlideIndex = (slideIndex) => {
    const offsetByIndex = this.slides.reduce((total, item, index) => {
      if (index >= slideIndex) {
        return total;
      }

      return total + (item.current.offsetWidth || 0);
    }, 0);

    return offsetByIndex;
  };

  getOffsetCorrectionForEdgeSlides = (activeSlideIndex, positionIndex) => {
    if (positionIndex - activeSlideIndex === 0) {
      return 0;
    }

    if (this.direction.toLowerCase() === 'forward' && activeSlideIndex < positionIndex) {
      return this.itemsListRef.current.offsetWidth / 3;
    }

    if (this.direction.toLowerCase() === 'backward' && activeSlideIndex > positionIndex) {
      return -this.itemsListRef.current.offsetWidth / 3;
    }

    return 0;
  }

  getSlideItemOnClick = ({
    index,
    activeSlideIndex,
    direction,
    onClick,
  }) => {
    const slideItemOnClick = (event) => {
      const forwardDirectionValue = activeSlideIndex < index ? 'forward' : '';
      const backwardDirectionValue = activeSlideIndex > index ? 'backward' : '';

      this.updateActiveSlideIndex(
        index,
        direction || forwardDirectionValue || backwardDirectionValue,
      );

      if (onClick) {
        onClick(event);
      }
    };

    return slideItemOnClick;
  }

  getNextSlideIndex = (direction) => {
    const { activeSlideIndex, itemsToScroll, children } = this.renderProps;

    const lastSlideIndex = Children.count(children) - 1;

    if (direction === 'forward') {
      const nextSlideIndex = activeSlideIndex + itemsToScroll;
      const isOnEnd = nextSlideIndex > lastSlideIndex;
      const newSlideIndex = isOnEnd ? nextSlideIndex - lastSlideIndex - 1 : nextSlideIndex;

      return newSlideIndex;
    }

    if (direction === 'backward') {
      const nextSlideIndex = activeSlideIndex - itemsToScroll;
      const isOnStart = nextSlideIndex < 0;
      const newSlideIndex = isOnStart ? lastSlideIndex + 1 + nextSlideIndex : nextSlideIndex;

      return newSlideIndex;
    }

    return activeSlideIndex;
  }

  updateActiveSlideIndex = (newActiveSlideIndex, direction) => {
    const {
      activeSlideIndex,
      onRequestChange,
      speed,
      delay,
      easing,
    } = this.renderProps;

    this.itemsListRef.current.style.transition = speed || delay
      ? `transform ${speed}ms ${easing} ${delay}ms`
      : null;

    if (newActiveSlideIndex !== activeSlideIndex) {
      this.stopAutoplay();

      this.direction = direction;
      onRequestChange(newActiveSlideIndex);
    } else {
      this.itemsListRef.current.style.transform = `translateX(-${this.itemsListRef.current.offsetWidth / 3}px)`;
    }
  }

  startAutoplay = () => {
    const {
      autoplay,
      autoplayDirection,
      delay,
    } = this.renderProps;

    if (autoplay) {
      this.autoplayTimer = setTimeout(() => {
        this.updateActiveSlideIndex(this.getNextSlideIndex(autoplayDirection), autoplayDirection);
      }, delay);
    }
  }

  stopAutoplay = () => {
    clearTimeout(this.autoplayTimer);
  }

  handleInitializationEnd = () => {
    this.setState(() => ({
      windowWidth: window.innerWidth,
    }), this.startAutoplay);
  };

  handleWindowResize = () => {
    clearTimeout(this.resizeTimer);

    this.resizeTimer = setTimeout(this.handleInitializationEnd, 400);
  };

  handleContainerClickCapture = (event) => {
    const { containerProps: { onClickCapture: containerOnClickCapture } } = this.renderProps;

    if (this.isListDragging) {
      event.preventDefault();
      event.stopPropagation();

      if (containerOnClickCapture) {
        containerOnClickCapture(event);
      }
    }
  }

  handleBackwardBtnClick = () => {
    this.updateActiveSlideIndex(this.getNextSlideIndex('backward'), 'backward');
  }

  handleForwardBtnClick = () => {
    this.updateActiveSlideIndex(this.getNextSlideIndex('forward'), 'forward');
  }

  handleItemsListTransitionEnd = () => {
    const { activeSlideIndex, onAfterChange } = this.renderProps;

    this.setState(() => ({
      positionIndex: activeSlideIndex,
    }), () => {
      const { positionIndex } = this.state;

      this.itemsListDragStartPos = null;
      this.isListDragging = false;

      this.startAutoplay();

      if (onAfterChange) {
        onAfterChange(activeSlideIndex, positionIndex);
      }
    });
  }

  updateItemsListPosByDragPos = (dragPos) => {
    const dragPosDiff = (this.itemsListDragStartPos - dragPos)
    + (this.itemsListRef.current.offsetWidth / 3);
    const minDragPos = 0;
    const maxDragPos = this.itemsListRef.current.offsetWidth - this.innerRef.current.offsetWidth;
    const itemsListPos = Math.max(Math.min(minDragPos, -dragPosDiff), -maxDragPos);

    this.itemsListRef.current.style.transition = 'none';
    this.itemsListRef.current.style.transform = `translateX(${itemsListPos}px)`;
  }

  handleItemsListDragEnd = (dragPos) => {
    const { activeSlideIndex } = this.renderProps;
    const mousePosDiff = this.itemsListDragStartPos - dragPos;
    const activeItemHalfWidth = this.slides[activeSlideIndex].current.offsetWidth / 2;

    if (mousePosDiff > activeItemHalfWidth) {
      this.updateActiveSlideIndex(this.getNextSlideIndex('forward'), 'forward');
    } else if (mousePosDiff < -activeItemHalfWidth) {
      this.updateActiveSlideIndex(this.getNextSlideIndex('backward'), 'backward');
    } else {
      this.updateActiveSlideIndex(activeSlideIndex, 'forward');
    }
  }

  handleItemsListMouseMove = (event) => {
    this.isListDragging = true;

    this.updateItemsListPosByDragPos(event.clientX);
  }

  handleItemsListMouseUp = (event) => {
    document.removeEventListener('mousemove', this.handleItemsListMouseMove);
    document.removeEventListener('mouseup', this.handleItemsListMouseUp);

    if (this.isListDragging) {
      this.handleItemsListDragEnd(event.clientX);
    }
  }

  handleItemsListMouseDown = (event) => {
    this.stopAutoplay();

    if (!this.isListDragging) {
      this.itemsListDragStartPos = event.clientX;

      document.addEventListener('mousemove', this.handleItemsListMouseMove);
      document.addEventListener('mouseup', this.handleItemsListMouseUp);
    }
  };

  handleItemsListTouchMove = (event) => {
    this.isListDragging = true;
    this.updateItemsListPosByDragPos(event.touches[0].clientX);
  };

  handleItemsListTouchEnd = (event) => {
    document.removeEventListener('touchmove', this.handleItemsListTouchMove);
    document.removeEventListener('touchend', this.handleItemsListTouchEnd);

    if (this.isListDragging) {
      this.handleItemsListDragEnd(event.changedTouches[event.changedTouches.length - 1].clientX);
    }
  };

  handleItemsListTouchStart = (event) => {
    this.stopAutoplay();

    if (!this.isListDragging) {
      this.itemsListDragStartPos = event.touches[0].clientX;

      document.addEventListener('touchmove', this.handleItemsListTouchMove);
      document.addEventListener('touchend', this.handleItemsListTouchEnd);
    }
  };

  renderSlidesItems = (items, startIndex) => {
    const {
      updateOnItemClick,
      activeSlideIndex,
      activeSlideProps: {
        className: activeSlideClassName = '',
        style: activeSlideStyle = {},
        ...activeSlideProps
      },
    } = this.renderProps;

    return (
      items.map((
        {
          props: {
            className: itemClassName = '',
            style: itemStyle = {},
            onClick: itemOnClick,
            ...itemComponentProps
          } = {},
          ...slideComponentData
        },
        index,
      ) => {
        const direction = this.renderedSlidesCount >= this.slidesCount ? 'forward' : 'backward';

        const isActive = index + startIndex === activeSlideIndex;

        const className = `${itemClassName} ${isActive ? activeSlideClassName : ''}`;
        const style = {
          ...itemStyle,
          ...(isActive ? activeSlideStyle : {}),
          boxSizing: 'border-box',
          margin: 0,
        };
        const onClick = updateOnItemClick
          ? this.getSlideItemOnClick({
            index: index + startIndex,
            direction,
            onClick: itemOnClick,
            activeSlideIndex,
          })
          : itemOnClick;
        const props = {
          className,
          style,
          onClick,
          ...itemComponentProps,
          ...(isActive ? activeSlideProps : {}),
        };

        this.slides[index + startIndex] = createRef();

        this.renderedSlidesCount = this.renderedSlidesCount + 1;

        return {
          props: {
            role: 'tabpanel',
            ...props,
          },
          ...slideComponentData,
          ref: this.slides[index + startIndex],
        };
      })
    );
  }

  render() {
    this.renderProps = this.getRenderProps(this.state, this.props);

    const {
      activeSlideIndex,
      speed,
      delay,
      easing,
      children,
      hideNavIfAllVisible,
      containerProps: {
        className: containerClassName = '',
        onClickCapture: containerOnClickCapture,
        ...containerProps
      },
      innerProps: {
        className: innerClassName = '',
        style: {
          width: innerWidthStyle = null,
          ...innerStyle
        } = {},
        ...innerProps
      },
      itemsListProps: {
        className: itemsListClassName = '',
        style: itemsListStyle = {},
        onTouchStart: onItemsListTouchStart,
        onMouseDown: onItemsListMouseDown,
        onTransitionEnd: onItemsListTransitionEnd,
        ...itemsListProps
      },
      forwardBtnProps: {
        children: forwardBtnChildren = null,
        show: showForwardBtn = true,
        ...forwardBtnProps
      },
      backwardBtnProps: {
        children: backwardBtnChildren = null,
        show: showBackwardBtn = true,
        ...backwardBtnProps
      },
    } = this.renderProps;

    const { windowWidth, positionIndex } = this.state;

    const slidesItems = Children.toArray(children);

    const innerWidth = this.getInnerWidth();

    const renderNavBtns = windowWidth && hideNavIfAllVisible
      ? (this.itemsListRef.current.offsetWidth / 3)
        > (innerWidth || this.innerRef.current.offsetWidth)
      : true;

    const isNewSLideIndex = (activeSlideIndex - positionIndex) !== 0;

    const positionIndexOffset = windowWidth && isNewSLideIndex
      ? this.getItemsListOffsetBySlideIndex(positionIndex)
      : 0;
    const activeSlideIndexOffset = windowWidth && isNewSLideIndex
      ? this.getItemsListOffsetBySlideIndex(activeSlideIndex)
      : 0;

    const itemsListTransition = !isNewSLideIndex || !(speed || delay)
      ? null
      : `transform ${speed}ms ${easing} ${delay}ms`;
    const itemsListTransform = windowWidth
      ? `translateX(-${(activeSlideIndexOffset - positionIndexOffset + this.getOffsetCorrectionForEdgeSlides(activeSlideIndex, positionIndex)) + this.itemsListRef.current.offsetWidth / 3}px)`
      : null;

    this.slides = [];
    this.slidesCount = slidesItems.length;
    this.renderedSlidesCount = 0;

    return (
      <div
        className={`${styles.ReactJSSimpleCarousel} ${containerClassName}`}
        onClickCapture={this.handleContainerClickCapture}
        {...containerProps}
        ref={this.containerRef}
      >
        {showBackwardBtn && renderNavBtns && (
          <button
            {...backwardBtnProps}
            type="button"
            onClick={this.handleBackwardBtnClick}
          >
            {backwardBtnChildren}
          </button>
        )}

        <div
          className={`${styles.ReactJSSimpleCarousel__inner} ${innerClassName}`}
          style={{
            width: innerWidth ? `${innerWidth}px` : innerWidthStyle,
            ...innerStyle,
          }}
          {...innerProps}
          ref={this.innerRef}
        >
          <div
            className={`${styles.ReactJSSimpleCarousel__itemsList} ${itemsListClassName}`}
            style={{
              ...itemsListStyle,
              transition: itemsListTransition,
              transform: itemsListTransform,
            }}
            onTouchStart={this.handleItemsListTouchStart}
            onMouseDown={this.handleItemsListMouseDown}
            onTransitionEnd={this.handleItemsListTransitionEnd}
            tabIndex="-1"
            role="presentation"
            {...itemsListProps}
            ref={this.itemsListRef}
          >

            {this.renderSlidesItems(slidesItems.slice(positionIndex), positionIndex)}
            {this.renderSlidesItems(slidesItems, 0)}
            {this.renderSlidesItems(slidesItems, 0)}
            {this.renderSlidesItems(slidesItems.slice(0, positionIndex), 0)}
          </div>
        </div>

        {showForwardBtn && renderNavBtns && (
          <button
            {...forwardBtnProps}
            type="button"
            onClick={this.handleForwardBtnClick}
          >
            {forwardBtnChildren}
          </button>
        )}
      </div>
    );
  }
}

export default ReactJSSimpleCarousel;
